;;   Delegate.scm - basic tests for gnustep classes having delegates
;;                  and targets and generally setting ivars
;;   Copyright (C) 2001 Free SoftwareFoundation, Inc.
;;
;;   Written by: Nicola Pero <n.pero@mi.flashnet.it>
;;   Date: October 2001, January 2002
;;
;;   This file is part of the GNUstep project.
;;
;;   This library is free software; you can redistribute it and/or
;;   modify it under the terms of the GNU Library General Public
;;   License as published by the Free Software Foundation; either
;;   version 2 of the License, or (at your option) any later version.
;;   
;;   This library is distributed in the hope that it will be useful,
;;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;   Library General Public License for more details.
;;
;;   You should have received a copy of the GNU Library General Public
;;   License along with this library; if not, write to the Free
;;   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;; You need to call delegate-init before using any of the procedures
;; in this file.  It will load Resources/IvarProxy.bundle, which
;; provides a little Objective-C class allowing us to perform some
;; memory checking to catch leaks in dealloc - memory checkings which
;; we can't do directly from guile (unfortunately, because of the
;; guile conservative garbage collector)
;;
;; the 'path' argument is the path from the directory in which you run
;; the tests to this directory.  In practice, it's './' for the base
;; tests, and '../' for the gui tests.
(define (init-Delegate-with-path path)
  (begin
    ([] ([] "NSBundle" bundleWithPath: 
	 ([] "NSBundle" pathForResource: ($$ "IvarProxy")
	  ofType: ($$ "bundle")
	  inDirectory: ([] ([] ([] ([] "NSFileManager" defaultManager) 
				currentDirectoryPath)
			    stringByAppendingPathComponent: ($$ path))
			stringByAppendingPathComponent: ($$ "Resources"))))
     load)
    )
)

;;
;; The following procedure checks that set-method / get-method set/remove
;; correctly a weak-ivar represented by object-arg
;;
;; If 'setSelector' is non "", it also checks that dealloc does not
;; release the object-arg.  We need the selector because this is done
;; via ObjC.
;;
;; As an example, see test-delegate below
;;
(define (test-weak-ivar class-name ivar-name 
			set-procedure get-procedure set-selector
			object-arg)

  ;; The instance we test
  (define anInstance ([] ([] class-name alloc) init))
  
  ;; Temp variable
  (define count 0)

  (greg-testcase 
   (string-append "Setting " ivar-name " in class '" class-name "' works") #t
   (lambda ()
     (set-procedure anInstance object-arg)
     (gstep-bool ([] object-arg isEqual: (get-procedure anInstance)))))
  
  (greg-testcase 
   (string-append "Setting " ivar-name " to nil "
		  "in class '" class-name "' works") 
   #t
   (lambda ()
     (set-procedure anInstance gstep-nil)
     (eq? (get-procedure anInstance) gstep-nil)))
  
  (greg-testcase
   (string-append "Setting " ivar-name " in class '" class-name  
		  "' does not retain it") 
   #t
   (lambda ()
     (set! count ([] object-arg retainCount))
     (set-procedure anInstance object-arg)
     (eq? count ([] object-arg retainCount))))

  (greg-testcase
   (string-append "Setting " ivar-name " in class '" class-name 
		  "' does not release the previous value") 
   #t
   (lambda ()
     (set! count ([] object-arg retainCount))
     (set-procedure anInstance object-arg)
     (set-procedure anInstance gstep-nil)
     (eq? count ([] object-arg retainCount))))


  (if (not (equal? set-selector ""))
      (greg-testcase
       (string-append "Dealloc in class '" class-name 
		      "' does not releases " ivar-name)
       #t
       (lambda ()
	 ;; Create a holding object
	 (define temp-instance ([] ([] "IvarProxy" alloc) 
			       initWithClass: ($$ class-name)))
	 ;; you can image such perverse situations that the following
	 ;; might prevent a crash
	 ([] object-arg retain)
	 ;; Save the retain count of the object now
	 (set! count ([] object-arg retainCount))
	 ;; Set the object into an ivar
	 ([] temp-instance setObjectIvar: object-arg 
	  usingSelector: set-selector)
	 ;; Destroy the holding object
	 ([] temp-instance deallocTheObject)
	 ;; The ivar count should be the same
	 (eq? count ([] object-arg retainCount)))))
)


;;
;; Basic test - check that -delegate, -setDelegate: works properly
;;
(define (test-delegate class-name)
  (test-weak-ivar 
   class-name 
   "delegate"
   (lambda (receiver arg) ([] receiver setDelegate: arg))
   (lambda (receiver) ([] receiver delegate))
   "setDelegate:"
   ([] "NSObject" new)))

;;
;; Basic test - check that -target, -setTarget: works properly
;;
(define (test-target class-name)
  (test-weak-ivar 
   class-name 
   "target"
   (lambda (receiver arg) ([] receiver setTarget: arg))
   (lambda (receiver) ([] receiver target))
   "setTarget:"
   ([] "NSObject" new)))


;;
;; Test for standard ivars which need to be retained/released appropriately
;;
(define (test-standard-ivar class-name ivar-name 
			    set-procedure get-procedure 
			    set-selector
			    object-arg)
  
  ;; The instance we test
  (define anInstance ([] ([] class-name alloc) init))
  
  ;; Temp variable
  (define count 0)

  (greg-testcase 
   (string-append "Setting " ivar-name " in class '" class-name "' works") #t
   (lambda ()
     (set-procedure anInstance object-arg)
     (gstep-bool ([] object-arg isEqual: (get-procedure anInstance)))))
  
  (greg-testcase 
   (string-append "Setting " ivar-name " to nil "
		  "in class '" class-name "' works") 
   #t
   (lambda ()
     (set-procedure anInstance gstep-nil)
     (eq? (get-procedure anInstance) gstep-nil)))
  
  (greg-testcase
   (string-append "Setting " ivar-name " in class '" class-name 
		  "' retains it") 
   #t
   (lambda ()
     (set! count ([] object-arg retainCount))
     (set-procedure anInstance object-arg)
     (eq? ([] object-arg retainCount) (+ count 1))))

  (greg-testcase
   (string-append "Setting " ivar-name " in class '" class-name 
		  "' releases the previous value") 
   #t
   (lambda ()
     (set-procedure anInstance gstep-nil)
     (set! count ([] object-arg retainCount))
     (set-procedure anInstance object-arg)
     (set-procedure anInstance gstep-nil)
     (eq? count ([] object-arg retainCount))))
  
  (if (not (equal? set-selector ""))
      (greg-testcase
       (string-append "Dealloc in class '" class-name 
		      "' releases " ivar-name)
       #t
       (lambda ()
	 ;; Create a holding object
	 (define temp-instance ([] ([] "IvarProxy" alloc) 
			       initWithClass: ($$ class-name)))
	 ;; you can image such perverse situations that the following
	 ;; might prevent a crash
	 ([] object-arg retain)
	 ;; Save the retain count of the object now
	 (set! count ([] object-arg retainCount))
	 ;; Set the object into an ivar
	 ([] temp-instance setObjectIvar: object-arg 
	  usingSelector: set-selector)
	 ;; Destroy the holding object
	 ([] temp-instance deallocTheObject)
	 ;; The ivar count should be the same
	 (eq? count ([] object-arg retainCount)))))
  )

  
