;;   Methods.scm - basic tests for gnustep methods from guile
;;   Copyright (C) 2002 Free SoftwareFoundation, Inc.
;;
;;   Written by:  Nicola Pero <nicola@brainstorm.co.uk>
;;   Date: February 2002
;;
;;   This file is part of the GNUstep project.
;;
;;   This library is free software;; you can redistribute it and/or
;;   modify it under the terms of the GNU Library General Public
;;   License as published by the Free Software Foundation;; either
;;   version 2 of the License, or (at your option) any later version.
;;   
;;   This library is distributed in the hope that it will be useful,
;;   but WITHOUT ANY WARRANTY;; without even the implied warranty of
;;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;   Library General Public License for more details.
;;
;;   You should have received a copy of the GNU Library General Public
;;   License along with this library;; if not, write to the Free
;;   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;
;; This file provides procedures to quickly write tests doing simply things
;; like checking that a method returns nil, or non-nil, or an object of a
;; certain class, or a non-empty string, or a non-empty array, or that two
;; objects are equal.
;;

;; test-for-nil checks that the result of executing 'code' is nil.

;;
;; Example usage:
;; (test-for-nil "NSBundle -initWithPath with invalid path returns nil" 
;;               ([] ([] "NSBundle" alloc) initWithPath: gstep-nil))
;;
(define (test-for-nil test-description code-which-should-return-nil)
  (greg-testcase 
   test-description #t
   (lambda ()
     (equal? code-which-should-return-nil gstep-nil)
     )
   )
  )

;; test-for-non-nil checks that the result of executing 'code' is not nil.

;;
;; Example usage:
;; (test-for-non nil "NSBundle -principalClass returns a valid class" 
;;                    ([] bundle principalClass))
;;
(define (test-for-non-nil test-description code-which-should-not-return-nil)
  (greg-testcase 
   test-description #t
   (lambda ()
     (not (equal? code-which-should-not-return-nil gstep-nil))
     )
   )
  )

;; test-for-string checks that the result of executing 'code' is a
;; non-null NSString with a non-0 length.

;;
;; Example usage:
;; (test-for-string "NSBundle -path returns a string" 
;;                  ([] bundle path))
;;
(define (test-for-string test-description code-which-should-return-a-string)
  (greg-testcase 
   test-description #t
   (lambda ()
     (define hopefully-a-string code-which-should-return-a-string)
     (and
      (not (equal? hopefully-a-string gstep-nil))
      (gstep-bool ([] hopefully-a-string isKindOfClass: ([] "NSString" class)))
      (not (eq? ([] hopefully-a-string length) 0))
      )
     )
   )
  )

;; test-for-array checks that the result of executing 'code' is a
;; non-null NSArray with a non-0 count.

;;
;; Example usage:
;; (test-for-array "NSBundle +allBundles returns a array" 
;;                 ([] "NSBundle" allBundles))
;;
(define (test-for-array test-description code-which-should-return-a-array)
  (greg-testcase 
   test-description #t
   (lambda ()
     (define hopefully-a-array code-which-should-return-a-array)
     (and
      (not (equal? hopefully-a-array gstep-nil))
      (gstep-bool ([] hopefully-a-array isKindOfClass: ([] "NSArray" class)))
      (not (eq? ([] hopefully-a-array count) 0))
      )
     )
   )
  )

;; test-for-class checks that the result of executing some code is an
;; instance of a certain class

;;
;; Example usage:
;; (test-for-class "NSBundle +bundleWithPath: returns a bundle" "NSBundle" 
;;                 ([] "NSBundle" bundleWithPath: ($$ "/usr")))
;;
(define (test-for-class test-description 
			class-name
			code-which-should-return-an-instance-of-that-class)
  (greg-testcase 
   test-description #t
   (lambda ()
     (define result code-which-should-return-an-instance-of-that-class)
     (and
      (not (equal? result gstep-nil))
      (gstep-bool ([] result isKindOfClass: ([] class-name class)))
      )
     )
   )
  )

;; test-for-equal-objects checks that object-a and object-b are equal,
;; as determined by calling ([] object-a isEqual: object-b) AND ([]
;; object-b isEqual: object-a)

;;
;; Example usage:
;;
;; (test-for-equal-objects "bundlePath returns the correct path"
;;			  ([] bundle bundlePath) path)
;;
(define (test-for-equal-objects test-description
				object-a
				object-b)
  (greg-testcase 
   test-description #t
   (lambda () 
     ;; Define internal objects to prevent object-a and object-b from
     ;; being evaluated twice.
     (define internal-object-a object-a)
     (define internal-object-b object-b)

     (and
      ;; The test passes only if both comparisons return true :-)
      ;; This is to make sure we don't get caught into silly tricks depending
      ;; on the order of the objects.
      (gstep-bool ([] internal-object-a isEqual: internal-object-b))
      (gstep-bool ([] internal-object-b isEqual: internal-object-a))
      )
     )))


