;   test00.scm - Tests for parsing simple property lists
;   Copyright (C) 2001 Free SoftwareFoundation, Inc.
;
;   Written by:  Nicola Pero <nicola@brainstorm.co.uk>
;   Date: May 2001
;
;   This file is part of the GNUstep project.
;
;   This library is free software; you can redistribute it and/or
;   modify it under the terms of the GNU Library General Public
;   License as published by the Free Software Foundation; either
;   version 2 of the License, or (at your option) any later version.
;   
;   This library is distributed in the hope that it will be useful,
;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;   Library General Public License for more details.
;
;   You should have received a copy of the GNU Library General Public
;   License along with this library; if not, write to the Free
;   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

(begin

  (define test-parse 
    (lambda (string result)
      (begin
	(gstep-bool ([] ([] ($$ string) propertyList) isEqual: result))
      )
    )
  )
  
  (greg-testcase 
   "We can parse a string" #t
   (lambda ()
     (test-parse "ariosto" ($$ "ariosto"))))

  (greg-testcase 
   "We can parse a quoted string" #t
   (lambda ()
     (test-parse "\"ariosto\"" ($$ "ariosto"))))

  (greg-testcase 
   "We can parse a quoted string with octal escape" #t
   (lambda ()
     (test-parse "\"ari\\033osto\"" ($$ "ariosto"))))

  (greg-testcase 
   "We can parse an empty array" #t
   (lambda ()
     (let ((result ([] "NSArray" array)))
       (test-parse "()" result)))) 

  (greg-testcase 
   "We can parse a simple array with a single object" #t
   (lambda ()
     (let ((result ([] "NSArray" arrayWithObject: ($$ "ariosto"))))
       (test-parse "(ariosto)" result)))) 

  (greg-testcase 
   "We can parse a simple array with a single object between \"s" #t
   (lambda ()
     (let ((result ([] "NSArray" arrayWithObject: ($$ "ariosto"))))
       (test-parse "(\"ariosto\")" result)))) 

  (greg-testcase 
   "We can parse a simple array with two objects" #t
   (lambda ()
     (let ((result gstep-nil))
       (set! result ([] "NSArray" arrayWithObject: ($$ "ariosto")))
       (set! result ([] result arrayByAddingObject: ($$ "boccaccio")))
       (test-parse "(ariosto, boccaccio)" result))))

  (greg-testcase 
   "We can parse a simple array with three objects, with \"s" #t
   (lambda ()
     (let ((result gstep-nil))
       (set! result ([] "NSArray" arrayWithObject: ($$ "ariosto")))
       (set! result ([] result arrayByAddingObject: ($$ "boccaccio")))
       (set! result ([] result arrayByAddingObject: ($$ "leopardi")))
       (test-parse "(ariosto, boccaccio, \"leopardi\")" result))))

  (greg-testcase 
   "We can parse a simple array with two objects and a C comment" #t
   (lambda ()
     (let ((result gstep-nil))
       (set! result ([] "NSArray" arrayWithObject: ($$ "ariosto")))
       (set! result ([] result arrayByAddingObject: ($$ "boccaccio")))
       (test-parse "(ariosto /* I like this one */, boccaccio)" result))))

  (greg-testcase 
   "We can parse a simple array with two objects and a C++ comment" #t
   (lambda ()
     (let ((result gstep-nil))
       (set! result ([] "NSArray" arrayWithObject: ($$ "ariosto")))
       (set! result ([] result arrayByAddingObject: ($$ "boccaccio")))
       (test-parse "(ariosto, // I like this one\n boccaccio)" result))))

  (greg-testcase 
   "We can parse an empty dictionary" #t
   (lambda ()
     (let ((result ([] "NSDictionary" dictionary)))
       (test-parse "{}" result)))) 

  (greg-testcase 
   "We can parse a simple dictionary with a single key value pair" #t
   (lambda ()
     (let ((result ([] "NSDictionary" dictionaryWithObject: ($$ "ariosto")
		    forKey: ($$ "author"))))
       (test-parse "{author = ariosto;}" result)))) 

  (greg-testcase 
   "We can parse a simple dictionary with a two key value pairs" #t
   (lambda ()
     (let ((result gstep-nil))
       (set! result ([] "NSMutableDictionary" dictionary))
       ([] result setObject: ($$ "ariosto") forKey: ($$ "author"))
       ([] result setObject: ($$ "orlando furioso") forKey: ($$ "title"))
       (test-parse "{author = ariosto; title = \"orlando furioso\"}" 
		   result))))

  (greg-testcase 
   "We can parse a simple dictionary with a two key value pairs and a C comment" #t
   (lambda ()
     (let ((result gstep-nil))
       (set! result ([] "NSMutableDictionary" dictionary))
       ([] result setObject: ($$ "ariosto") forKey: ($$ "author"))
       ([] result setObject: ($$ "orlando furioso") forKey: ($$ "title"))
       (test-parse "{/* -*-c-*- */ author = ariosto; title = \"orlando furioso\"}" 
		   result))))

  (greg-testcase 
   "We can parse a simple dictionary with a two key value pairs and a C++ comment" #t
   (lambda ()
     (let ((result gstep-nil))
       (set! result ([] "NSMutableDictionary" dictionary))
       ([] result setObject: ($$ "ariosto") forKey: ($$ "author"))
       ([] result setObject: ($$ "orlando furioso") forKey: ($$ "title"))
       (test-parse "{// -*-c-*-\n author = ariosto; title = \"orlando furioso\"}" 
		   result))))

  (greg-testcase 
   "We can parse an array containing a single array" #t
   (lambda ()
     (let ((result ([] "NSArray" arrayWithObject: 
		    ([] "NSArray" arrayWithObject: ($$ "ariosto")))))
	   (test-parse "((ariosto))" result))))

  (greg-testcase 
   "We can parse an array containing a single dictionary" #t
   (lambda ()
     (let ((result ([] "NSArray" arrayWithObject: 
		    ([] "NSDictionary" dictionaryWithObject: ($$ "ariosto")
		     forKey: ($$ "author")))))
       (test-parse "({author = ariosto;})" result))))

  (greg-testcase 
   "We can parse an array containing an array and a string" #t
   (lambda ()
     (let ((result gstep-nil))
       (set! result ([] "NSArray" arrayWithObject: 
		     ([] "NSArray" arrayWithObject: ($$ "farinata"))))
       (set! result ([] result arrayByAddingObject: ($$ "dante")))
       (test-parse "((farinata), dante)" result))))

  (greg-testcase 
   "We can parse an array containing an array and a dictionary" #t
   (lambda ()
     (let ((result gstep-nil))
       (set! result ([] "NSArray" arrayWithObject: 
		     ([] "NSArray" arrayWithObject: ($$ "farinata"))))
       (set! result ([] result arrayByAddingObject: 
		     ([] "NSDictionary" dictionaryWithObject: ($$ "ariosto")
		      forKey: ($$ "author"))))
       (test-parse "((farinata), {author = ariosto;})" result))))
  
  )


