;   Protocols.scm - Tests of GNUstep libraries from Guile
;   Copyright (C) 1998 Free SoftwareFoundation, Inc.
;
;   Written by:  Richard Frith-Macdonald <richard@brainstorm.co.uk>
;   Date: July 1998
;
;   This file is part of the GNUstep project.
;
;   This library is free software; you can redistribute it and/or
;   modify it under the terms of the GNU Library General Public
;   License as published by the Free Software Foundation; either
;   version 2 of the License, or (at your option) any later version.
;   
;   This library is distributed in the hope that it will be useful,
;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;   Library General Public License for more details.
;
;   You should have received a copy of the GNU Library General Public
;   License along with this library; if not, write to the Free
;   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;
;	This file declares procedures to test conformance to some of
;	the standard formal protocols.
;
(define	NSCodingProtocol (gstep-lookup-protocol "NSCoding"))
(define	NSCopyingProtocol (gstep-lookup-protocol "NSCopying"))
(define	NSMutableCopyingProtocol (gstep-lookup-protocol "NSMutableCopying"))
(define	NSObjectProtocol (gstep-lookup-protocol "NSObject"))

;
;	Define a procedure to perform tests of the NSCoding protocol
;	Expects a single argument - a list of objects to be tested.
;
(define (test-NSCoding objects)
  (do
    ((list objects (cdr list)) (counter 1 (+ counter 1)))
    ((eq? list '()))
    (let*
      (
	(theObj (car list))
	(className  "")
	(prefix "")
	(conforms #f)
	(archiver gstep-nil)
	(myData gstep-nil)
	(decoded gstep-nil)
      )
      (begin
	;
	;  Get the description/name of the class of object being tested.
	;
	(greg-testcase
	  (string-append "I can extract a class name for object "
	    (number->string counter)) #t
	  (lambda ()
	    (set! className ([] ([] ([] theObj class) description) cString))
	    (not (equal? className ""))
	  )
	)

	;
	; Set up prefix for test assertions.
	;
	(set! prefix (string-append "Object " (number->string counter)
	  " of class '" className "' "))

	;
	; Check protocol conformance then proceed with tests.
	;
	(greg-testcase (string-append prefix "conforms to NSCoding") #t
	  (lambda ()
	    (set! conforms
	      (gstep-bool->bool
		([] theObj conformsToProtocol: NSCodingProtocol))
	    )
	    conforms
	  )
	)

	(if conforms
	  (begin
	    (greg-testcase " - I am able to set up an archiver" #t
	      (lambda ()
		(set! myData ([] "NSMutableData" data))
		(set! archiver ([] ([] "NSArchiver" alloc)
		      initForWritingWithMutableData: myData))
		(gc)
		(not (equal? archiver gstep-nil))
	      )
	    )

	    (set! myData gstep-nil)
	    (greg-testcase
	      (string-append prefix "can be encoded") #t
	      (lambda ()
		([] archiver encodeRootObject: theObj)
		(set! myData ([] archiver archiverData))
		(> ([] myData length) 0)
	      )
	    )

	    (greg-testcase
	      (string-append prefix "can be decoded") #t
	      (lambda ()
		(set! decoded
		  ([] "NSUnarchiver" unarchiveObjectWithData: myData))
		(not (equal? decoded gstep-nil))
	      )
	    )

	    (greg-testcase
	      (string-append prefix "decoded object equals original") #t
	      (lambda ()
		(gstep-bool->bool ([] theObj isEqual: decoded))
	      )
	    )

	    (gc)
	  )
	)
      )
    )
  )
)


;
;	This procedure must be invoked with five arguments -
;	    A string naming the immutable variant of the class
;	    A string naming the mutable variant of the class
;	    A list of objects to be tested.
;	    A flag to say whether a copy of an immutable object should be a
;	    simple retain irrespective of zones.
;	    A flag to say whether a copy of an immutable object should be a
;	    true copy irrespective of zones.
;
;	The procedure will test the NSCopying protocol PLUS the
;	[-copy] method (inherited from NSObject).
;
(define (test-NSCopying iClassName mClassName objects must-retain must-copy)

  (define iClass gstep-nil)
  (define mClass gstep-nil)

  ;
  ; Make a note of the class we expect to end up with after a copy
  ;
  (greg-testcase (string-append iClassName " is a known class") #t
    (lambda ()
      (set! iClass (gstep-lookup-class iClassName))
      (not (equal? iClass gstep-nil))
    )
  )

  ;
  ; Make a note of the class we expect to end up with after a mutable copy
  ;
  (greg-testcase (string-append mClassName " is a known class") #t
    (lambda ()
      (set! mClass (gstep-lookup-class mClassName))
      (not (equal? iClass gstep-nil))
    )
  )

  (do
    ((list objects (cdr list)) (counter 1 (+ counter 1)))
    ((eq? list '()))
    (let*
      (
	(theObj (car list))
	(theName "")
	(theClass gstep-nil)
	(immutable #t)
	(prefix "")
	(conforms #f)
	(theCopy gstep-nil)
	(should-retain '())
      )
      (begin
	;
	; Check to see if the object to be copied is mutable or not.
	; NB. If the two classes are identical - there is no mutable version
	; of the class, so our object must be immutable.
	;
	(if (and (not (equal? iClass mClass))
	    (gstep-bool->bool ([] theObj isKindOfClass: mClass)))
	  (set! immutable #f)
	  (set! immutable #t)
	)

	;
	; Set up the name and class of the object to be created and copied.
	;
	(if immutable
	  (begin
	    (set! theName iClassName)
	    (set! theClass iClass)
	  )
	  (begin
	    (set! theName mClassName)
	    (set! theClass mClass)
	  )
	)

	;
	; Set up prefix for test assertions.
	;
	(set! prefix (string-append "Object " (number->string counter)
	  " of class '" theName "' "))

	;
	; Check protocol conformance then proceed with tests.
	;
	(greg-testcase (string-append prefix "conforms to NSCopying") #t
	  (lambda ()
	    (set! conforms
	      (gstep-bool->bool
		([] theObj conformsToProtocol: NSCopyingProtocol))
	    )
	    conforms
	  )
	)

	(if conforms
	  (begin
	    (set! theCopy gstep-nil)
	    (set! should-retain '())
	    (if must-copy
	      (set! should-retain #f)
	      (if must-retain
		(set! should-retain #t)
		(set! should-retain (NSShouldRetainWithZone theObj defZone))
	      )
	    )

	    (greg-testcase (string-append prefix "understands [-copy]") #t
	      (lambda ()
		(set! theCopy ([] theObj copy))
		(not (equal? theCopy gstep-nil))
	      )
	    )

	    (greg-testcase (string-append prefix "copy is of correct type") #t
	      (lambda ()
		(gstep-bool->bool ([] theCopy isKindOfClass: iClass))
	      )
	    )

	    (greg-testcase (string-append prefix "[original isEqual: copy]") #t
	      (lambda ()
		(gstep-bool->bool ([] theObj isEqual: theCopy))
	      )
	    )

	    ;
	    ; If the class is immutable, then a copy within the same zone should 
	    ; actually be a simple retain, while a copy between zones should
	    ; result in a real copy.
	    ;
	    ; If the class is mutable, the new copy should not be.
	    ;
	    (if immutable
	      (if should-retain
		(greg-testcase
		  (string-append iClassName
		    " is retained by copy with same zone") #t
		  (lambda ()
		    (equal? theCopy theObj)
		  )
		)
		(greg-testcase
		  (string-append iClassName
		    " not retained by copy with same zone") #t
		  (lambda ()
		    (not (equal? theCopy theObj))
		  )
		)
	      )
	    )

	    (if (not (equal? theClass iClass))
	      (greg-testcase
		(string-append prefix "result of copy is not immutable") #t
		(lambda ()
		  (not (gstep-bool->bool
			([] theCopy isKindOfClass: theClass)))
		)
	      )
	    )

	    ;
	    ; Now do it with another zone.
	    ;
	    (if must-copy
	      (set! should-retain #f)
	      (if must-retain
		(set! should-retain #t)
		(set! should-retain (NSShouldRetainWithZone theObj testZone))
	      )
	    )

	    (greg-testcase
	      (string-append iClassName " understands [-copyWithZone:]") #t
	      (lambda ()
		(set! theCopy ([] theObj copyWithZone: testZone))
		(not (equal? theCopy gstep-nil))
	      )
	    )

	    (greg-testcase (string-append prefix "zCopy has correct type") #t
	      (lambda ()
		(gstep-bool->bool ([] theCopy isKindOfClass: iClass))
	      )
	    )

	    (greg-testcase (string-append prefix "[original isEqual: zCopy]") #t
	      (lambda ()
		(gstep-bool->bool ([] theObj isEqual: theCopy))
	      )
	    )

	    (if immutable
	      (if should-retain
		(greg-testcase
		  (string-append iClassName
		    " is retained by copy with other zone") #t
		  (lambda ()
		    (equal? theCopy theObj)
		  )
		)
		(greg-testcase
		  (string-append iClassName
		    " not retained by copy with other zone") #t
		  (lambda ()
		    (not (equal? theCopy theObj))
		  )
		)
	      )
	    )

	    (if (not (equal? theClass iClass))
	      (greg-testcase (string-append prefix
		"result of copyWithZone: is not immutable") #t
		(lambda ()
		  (not (gstep-bool->bool
			([] theCopy isKindOfClass: theClass)))
		)
	      )
	    )
	  )
	)
      )
    )
  )
)

;
;	This procedure must be invoked with three arguments -
;	    A string naming the immutable variant of the class
;	    A string naming the mutable variant of the class
;	    A list of objects (instance of either the mutable or the
;	    immutable class) to be copied.
;
;	The procedure will test the NSMutableCopying protocol PLUS the
;	[-mutableCopy] method (inherited from NSObject).
;
(define (test-NSMutableCopying iClassName mClassName objects)

  (define iClass gstep-nil)
  (define mClass gstep-nil)

  ;
  ; Make a note of the class we expect to end up with after a copy
  ;
  (greg-testcase (string-append iClassName " is a known class") #t
    (lambda ()
      (set! iClass (gstep-lookup-class iClassName))
      (not (equal? iClass gstep-nil))
    )
  )

  ;
  ; Make a note of the class we expect to end up with after a mutable copy
  ;
  (greg-testcase (string-append mClassName " is a known class") #t
    (lambda ()
      (set! mClass (gstep-lookup-class mClassName))
      (not (equal? iClass gstep-nil))
    )
  )

  (do
    ((list objects (cdr list)) (counter 1 (+ counter 1)))
    ((eq? list '()))
    (let*
      (
	(theObj (car list))
	(theName "")
	(theClass gstep-nil)
	(immutable #t)
	(prefix "")
	(conforms #f)
	(theCopy gstep-nil)
      )
      (begin
	;
	; Check to see if the object to be copied is mutable or not.
	; NB. If the two classes are identical - there is no mutable version
	; of the class, so our object must be immutable.
	;
	(if (and (not (equal? iClass mClass))
	    (gstep-bool->bool ([] theObj isKindOfClass: mClass)))
	  (set! immutable #f)
	  (set! immutable #t)
	)

	;
	; Set up the name and class of the object to be created and copied.
	;
	(if immutable
	  (begin
	    (set! theName iClassName)
	    (set! theClass iClass)
	  )
	  (begin
	    (set! theName mClassName)
	    (set! theClass mClass)
	  )
	)

	;
	; Set up prefix for test assertions.
	;
	(set! prefix (string-append "Object " (number->string counter)
	  " of class '" theName "' "))

	;
	; Check protocol conformance then proceed with tests.
	;
	(greg-testcase (string-append prefix "conforms to NSMutableCopying") #t
	  (lambda ()
	    (set! conforms
	      (gstep-bool->bool
		([] theObj conformsToProtocol: NSMutableCopyingProtocol))
	    )
	    conforms
	  )
	)

	(if conforms
	  (begin
	    (set! theCopy gstep-nil)
	    (greg-testcase
	      (string-append prefix "understands [-mutableCopy]") #t
	      (lambda ()
		(set! theCopy ([] theObj mutableCopy))
		(not (equal? theCopy gstep-nil))
	      )
	    )

	    (greg-testcase
	      (string-append prefix "mutable copy is of correct type") #t
	      (lambda ()
		(gstep-bool->bool ([] theCopy isKindOfClass: mClass))
	      )
	    )

	    (greg-testcase
	      (string-append prefix "[original isEqual: mutablecopy]") #t
	      (lambda ()
		(gstep-bool->bool ([] theObj isEqual: theCopy))
	      )
	    )

	    (greg-testcase
	      (string-append mClassName
		" not retained by mutable copy in same zone") #t
	      (lambda ()
		(not (equal? theCopy theObj))
	      )
	    )

	    (greg-testcase
	      (string-append
		mClassName " understands [-mutableCopyWithZone:]") #t
	      (lambda ()
		(set! theCopy ([] theObj mutableCopyWithZone: testZone))
		(not (equal? theCopy gstep-nil))
	      )
	    )

	    (greg-testcase
	      (string-append mClassName
		" not retained by mutable copy in other zone") #t
	      (lambda ()
		(not (equal? theCopy theObj))
	      )
	    )
	  )
	)
      )
    )
  )
)


;
;	Define a procedure to perform tests of the NSObject protocol
;	Expects two arguments - the name of the class of the objects
;	being tested (all the objects must be instances of the same
;	class) and a list of objects to test.
;
(define (test-NSObject className objects)

  ;
  ;  Check that we have a valid class name.
  ;
  (define theClass gstep-nil)
  (greg-testcase (string-append className " is a known class") #t
    (lambda ()
      (set! theClass (gstep-lookup-class className))
      (not (equal? theClass gstep-nil))
    )
  )

  (do
    ((list objects (cdr list)) (counter 1 (+ counter 1)))
    ((eq? list '()))
    (let*
      (
	(theObj (car list))
	(prefix "")
	(conforms #f)
	(mySelf gstep-nil)
	(myClass gstep-nil)
	(count0 gstep-nil)
	(count1 gstep-nil)
	(count2 gstep-nil)
      )
      (begin
	;
	; Set up prefix for test assertions.
	;
	(set! prefix (string-append "Object " (number->string counter)
	  " of class '" className "' "))

	;
	; Check protocol conformance then proceed with tests.
	;
	(greg-testcase (string-append prefix "conforms to NSObject") #t
	  (lambda ()
	    (set! conforms
	      (gstep-bool->bool
		([] theObj conformsToProtocol: NSObjectProtocol))
	    )
	    conforms
	  )
	)

	(if conforms
	  (begin
	    (set! mySelf gstep-nil)
	    (greg-testcase (string-append prefix "can return self") #t
	      (lambda ()
		(set! mySelf ([] theObj self))
		(equal? mySelf theObj)
	      )
	    )

	    (set! myClass gstep-nil)
	    (greg-testcase (string-append prefix "can return own class") #t
	      (lambda ()
		(set! myClass ([] theObj class))
		(not (equal? myClass gstep-nil))
	      )
	    )

	    (greg-testcase (string-append prefix "object is correct class") #t
	      (lambda ()
		(gstep-bool->bool ([] theObj isKindOfClass: theClass))
	      )
	    )

	    (greg-testcase (string-append prefix "isMemberOfClass works") #t
	      (lambda ()
		(gstep-bool->bool
		  (if (equal? mySelf myClass)
		    (not ([] theObj isMemberOfClass: myClass))
		    ([] theObj isMemberOfClass: myClass)
		  )
		)
	      )
	    )

	    ;
	    ; An NSObject has a superclass of 'nil'
	    ; Everything else should have a non-nil superclass that is not
	    ; the same as it's own class.
	    ;
	    (greg-testcase (string-append prefix "can return superclass") #t
	      (lambda ()
		(define sup ([] theObj superclass))
		(if (equal? theClass (gstep-lookup-class "NSObject"))
		  (equal? gstep-nil sup)
		  (and
		    (not (equal? gstep-nil sup))
		    (not (equal? myClass sup))
		  )
		)
	      )
	    )

	    (greg-testcase (string-append prefix "returns a hash value") #t
	      (lambda ()
		(number? ([] theObj hash))
	      )
	    )

	    (greg-testcase (string-append prefix "isEqual: to self") #t
	      (lambda ()
		(gstep-bool->bool ([] theObj isEqual: theObj))
	      )
	    )

	    (greg-testcase (string-append prefix "respondsToSelector:") #t
	      (lambda ()
		(gstep-bool->bool ([] theObj respondsToSelector: "self"))
	      )
	    )

	    (greg-testcase (string-append prefix "knows if it is a proxy") #t
	      (lambda ()
		([] theObj isProxy)
		#t
	      )
	    )

	    (greg-testcase (string-append prefix "handles retain") #t
	      (lambda ()
		(define r ([] theObj retain))
		(equal? theObj r)
	      )
	    )

	    (greg-testcase (string-append prefix "handles release") #t
	      (lambda ()
		([] theObj release)
		(gc)
		#t
	      )
	    )

	    (greg-testcase (string-append prefix "handles autorelease") #t
	      (lambda ()
		([] theObj retain)
		([] theObj autorelease)
		(gc)
		#t
	      )
	    )

	    (greg-testcase (string-append prefix "has working retainCount") #t
	      (lambda ()
		(gc)
		(set! count0 ([] theObj retainCount))
		([] theObj retain)
		(set! count1 ([] theObj retainCount))
		([] theObj release)
		(gc)
		(set! count2 ([] theObj retainCount))
		(= count0 count2)
	      )
	    )

	    (greg-testcase (string-append prefix "has NSString description") #t
	      (lambda ()
		(gstep-bool->bool
		  ([] ([] theObj description) isKindOfClass:
		    ([] "NSString" class)))
	      )
	    )

	    (greg-testcase (string-append prefix "handles performSelector:") #t
	      (lambda ()
		(define obj ([] theObj performSelector: "self"))
		(equal? obj theObj)
	      )
	    )

	    (greg-dlog prefix
	      "would like to test [-performSelector:withObject:] and"
	      " [-performSelector:withObject:withObject:], but we don't know"
	      " of any handy selectors (for methods returning ids) to test it"
	      " with\n")

	    (gc)
	  )
	)
      )
    )
  )
)

